package at.ac.tuwien.infosys.jaxb;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.xml.bind.annotation.Facets;
import javax.xml.bind.annotation.MaxOccurs;
import javax.xml.bind.annotation.MinOccurs;
import javax.xml.bind.annotation.Facets.FacetDefinition;
import javax.xml.namespace.QName;

import org.apache.log4j.Logger;

import com.sun.xml.bind.v2.model.core.PropertyInfo;
import com.sun.xml.bind.v2.model.core.TypeRef;
import com.sun.xml.bind.v2.schemagen.xmlschema.LocalElement;
import com.sun.xml.txw2.TypedXmlWriter;

/**
 * @author Waldemar Hummer (hummer@infosys.tuwien.ac.at)
 */
public class XmlSchemaEnhancer {

	public static final String NS_XSD = "http://www.w3.org/2001/XMLSchema";
	public static final Logger logger = Logger.getLogger(XmlSchemaEnhancer.class);
	
	public static <T,C> void addFacets(TypeRef<T,C> t, LocalElement e) {
		
		if(!hasFacets(t))
			return;
		
		Map<String,List<String>> facets = new HashMap<String, List<String>>();

		Facets facetsAnno = getFacetsAnnotation(t);
		
		try {
			facets = getDefinedFacets(facetsAnno);
		} catch (Exception ex) {
			logger.warn("Unable to add XSD Facets in Schema generated by JAXB.", ex);
			return;
		}
		
		TypedXmlWriter restriction = null;
		for(String facetName : facets.keySet()) {
			restriction = getRestriction(t, e, restriction);
			for(String facetValue : facets.get(facetName)) {
				restriction._element(new QName(NS_XSD,facetName), TypedXmlWriter.class).
						_attribute("value", facetValue);
			}
		}
		
	}
	
	private static <T,C> Facets getFacetsAnnotation(TypeRef<T,C> t) {
		if(!t.getTarget().isSimpleType())
			return null;
		
		try {
			Object value = getAnnotationOfProperty(t.getSource(), Facets.class);
			if(value instanceof Facets)
				return (Facets)value;
		} catch (Exception e2) {
			logger.warn("Unable to get Facets annotation from type " + t, e2);
		}
		return null;
	}
	
	public static <T,C> boolean hasFacets(TypeRef<T,C> t) {
		Facets facets = getFacetsAnnotation(t);
		
		if(facets == null)
			return false;
		
		try {
			Map<String,List<String>> definedFacets = getDefinedFacets(facets);
			return definedFacets.size() > 0;
		} catch (Exception ex) {
			logger.warn("Unable to get defined XSD Facets from type " + t, ex);
			return false;
		}
	}
	
	private static <T,C> Object getAnnotationOfProperty(PropertyInfo<T,C> info, Class<?> annoClass) throws Exception {
		if(annoClass == Facets.class && info.hasAnnotation(Facets.class))
			return info.readAnnotation(Facets.class);
		if(info.parent() == null)
			return null;
		if(!(info.parent().getType() instanceof Class<?>))
			return null;
		return getAnnotationOfProperty((Class<?>)info.parent().getType(), info.getName(), annoClass);
	}

	@SuppressWarnings("unchecked")
	private static Object getAnnotationOfProperty(
			Class<?> parent, String fieldName, Class annoClass) throws Exception {
		try {
			Field f = parent.getDeclaredField(fieldName);
			if(f == null)
				return null;
			return f.getAnnotation(annoClass);			
		} catch (Exception e) {
			throw new RuntimeException("Could not get annotation '" + annoClass.getSimpleName() + 
					"' of field " + fieldName + " of class " + parent);
		}
	}

	public static <T,C> boolean writeCustomOccurs(TypeRef<T,C> t, LocalElement e, boolean isOptional, boolean repeated) {
		
		MaxOccurs max = null;
		MinOccurs min = null;
		try {
			max = (MaxOccurs)getAnnotationOfProperty(t.getSource(), MaxOccurs.class);
		} catch (Exception e2) {
			logger.warn("Unable to get @MaxOccurs annotation from type " + t, e2);
		}
		try {
			min = (MinOccurs)getAnnotationOfProperty(t.getSource(), MinOccurs.class);
		} catch (Exception e2) {
			logger.warn("Unable to get @MinOccurs annotation from type " + t, e2);
		}
		
		if(min == null && max == null)
			return false;

		if(min != null){
			int value = (int)min.value();
			e.minOccurs(value);
		} else if(isOptional) {
			e.minOccurs(0);
		}
		
		if(max != null){
			int value = (int)max.value();
			e.maxOccurs(value);
		} else if(repeated) {
			e.maxOccurs("unbounded");
		}
		
		return true;
	}
	
	
	private static <T,C> TypedXmlWriter getRestriction(TypeRef<T,C> t, LocalElement e, TypedXmlWriter w) {
		if(w == null) {
			TypedXmlWriter st = e._element(new QName(NS_XSD,"simpleType"), TypedXmlWriter.class);
			TypedXmlWriter r = st._element(new QName(NS_XSD,"restriction"), TypedXmlWriter.class);
			r._attribute("base", t.getTarget().getTypeName());
			return r;
		}
		return w;
	}
	
	private static Map<String,List<String>> getDefinedFacets(Facets facetsAnnotation) throws Exception {
		List<Method> annoMethods = new LinkedList<Method>();
		Map<String,List<String>> result = new HashMap<String, List<String>>();
		
		if(facetsAnnotation == null)
			return result;
		
		for(Method m : Facets.class.getDeclaredMethods()) {
			if(m.isAnnotationPresent(FacetDefinition.class))
				annoMethods.add(m);
		}
		
		for(Method m : annoMethods) {
			Object value = m.invoke(facetsAnnotation);
			Object defaultValue = m.getDefaultValue();
			String facetName = m.getName();
			if(value != null && !value.equals(defaultValue)) {
				
				if(!result.containsKey(facetName)) {
					result.put(facetName, new LinkedList<String>());
				}
				if(value instanceof String[]) {
					for(String s : (String[])value)
						result.get(facetName).add(s);
				} else {
					result.get(facetName).add("" + value);
				}
			}
		}
		
		return result;
	}
}
